#!/usr/bin/env python3
import argparse

import matplotlib.patheffects as path_effects
import matplotlib.pyplot as plt
import matplotlib.ticker as tck
from DRexParam import dim, gridCoords, name
from numpy import linspace, load


def makeFilledContourPlot(
    mplAxis, xCoords, yCoords, data, levels, cmap, cbarTicks, cbarLab
):
    ctrf = mplAxis.contourf(
        xCoords, yCoords, data, antialiased=True, levels=levels, cmap=cmap
    )
    for iso in ctrf.collections:
        iso.set_edgecolor("face")
    mplAxis.set_aspect("equal", adjustable="box")
    mplAxis.set_ylabel(
        "Depth (km)",
        fontsize=18,
        fontweight="semibold",
        labelpad=20,
        color="xkcd:white",
        bbox=dict(boxstyle="round", facecolor="xkcd:black"),
    )
    mplAxis.xaxis.set_minor_locator(tck.AutoMinorLocator(2))
    mplAxis.yaxis.set_minor_locator(tck.AutoMinorLocator(2))
    mplAxis.tick_params(which="major", length=7, labelsize=16, width=2)
    mplAxis.tick_params(which="minor", length=4, width=2, color="xkcd:bright red")
    cbar = fig.colorbar(ctrf, ax=mplAxis, ticks=cbarTicks, fraction=0.05, aspect=10)
    cbar.set_label(
        cbarLab,
        labelpad=-25,
        fontweight="semibold",
        fontsize=18,
        color="xkcd:white",
        path_effects=[
            path_effects.Stroke(linewidth=3, foreground="black"),
            path_effects.Normal(),
        ],
    )
    cbar.ax.tick_params(which="major", labelsize=16, width=2, length=6)
    cbar.ax.tick_params(
        which="minor", labelsize=16, width=2, length=3, color="xkcd:bright red"
    )
    cbar.ax.yaxis.set_label_position("left")
    cbar.ax.yaxis.set_ticks_position("right")
    return mplAxis


parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description="""
A simple plotting script using Matplotlib to generate filled contour plots.
Requires the input file DRexParam.py in the same directory.""",
)
parser.add_argument("input", help="input file (expects a Numpy NpzFile)")
args = parser.parse_args()

DRexArr = load(args.input)

fig, (ax, bx) = plt.subplots(
    nrows=2, ncols=1, sharex=True, sharey=True, num=0, figsize=(20, 10)
)
ax = makeFilledContourPlot(
    ax,
    gridCoords[0][:-1] / 1e3,
    gridCoords[-1][:-1] / 1e3,
    DRexArr["percani"] if dim == 2 else DRexArr["percani"][:, :, 0],
    linspace(0, 7, 22),
    "inferno",
    linspace(0, 7, 8),
    "Azimuthal Anisotropy (%)",
)
bx = makeFilledContourPlot(
    bx,
    gridCoords[0][:-1] / 1e3,
    gridCoords[-1][:-1] / 1e3,
    DRexArr["radani"] if dim == 2 else DRexArr["radani"][:, :, 0],
    linspace(0.85, 1.15, 31),
    "RdBu",
    linspace(0.85, 1.15, 7),
    "Radial Anisotropy",
)

ax.invert_yaxis()
bx.set_xlabel(
    "Y Axis (km)",
    fontsize=18,
    fontweight="semibold",
    labelpad=15,
    color="xkcd:white",
    bbox=dict(boxstyle="round", facecolor="xkcd:black"),
)
fig.tight_layout()
fig.savefig(f"PyDRex{dim}D_{name}_Figure.pdf", bbox_inches="tight", dpi=500)



"""> PyDRex: Tensor operation functions and helpers.

For Voigt notation, the symmetric 6x6 matrix representation is used,
which assumes that the fourth order tensor being represented as such is also symmetric.
The vectorial notation uses 21 components which are the independent components of the
symmetric 6x6 matrix.

"""

import numba as nb
import numpy as np

PERMUTATION_SYMBOL = np.array(
    [
        [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, -1.0, 0.0]],
        [[0.0, 0.0, -1.0], [0.0, 0.0, 0.0], [1.0, 0.0, 0.0]],
        [[0.0, 1.0, 0.0], [-1.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
    ]
)


@nb.njit(fastmath=True)
def voigt_decompose(matrix):
    """Decompose elastic tensor (as 6x6 Voigt matrix) into distinct contractions.

    Return the only two independent contractions of the elastic tensor given as a 6x6
    Voigt `matrix`. With reference to the full 4-th order elastic tensor, the
    contractions are defined as:
    - $d_{ij} = C_{ijkk}$ (dilatational stiffness tensor)
    - $v_{ij} = C_{ijkj}$ (deviatoric stiffness tensor)

    Any vector which is an eigenvector of both $d_{ij}$ and $v_{ij}$ is always normal to
    a symmetry plane of the elastic medium.

    See Equations 3.4 & 3.5 in [Browaeys & Chevrot (2004)](https://doi.org/10.1111/j.1365-246X.2004.02415.x).

    """
    # 1. Compose dᵢⱼ = Cᵢⱼₖₖ (dilatational stiffness tensor)
    # Eq. 3.4 in Browaeys & Chevrot, 2004.
    stiffness_dilat = np.empty((3, 3))
    for i in range(3):
        stiffness_dilat[i, i] = matrix[:3, i].sum()
    stiffness_dilat[0, 1] = stiffness_dilat[1, 0] = matrix[:3, 5].sum()
    stiffness_dilat[0, 2] = stiffness_dilat[2, 0] = matrix[:3, 4].sum()
    stiffness_dilat[1, 2] = stiffness_dilat[2, 1] = matrix[:3, 3].sum()
    # 2. Compose vᵢⱼ = Cᵢⱼₖⱼ (deviatoric stiffness tensor)
    # Eq. 3.5, Browaeys & Chevrot, 2004.
    stiffness_deviat = np.empty((3, 3))
    stiffness_deviat[0, 0] = matrix[0, 0] + matrix[4, 4] + matrix[5, 5]
    stiffness_deviat[1, 1] = matrix[1, 1] + matrix[3, 3] + matrix[5, 5]
    stiffness_deviat[2, 2] = matrix[2, 2] + matrix[3, 3] + matrix[4, 4]
    stiffness_deviat[0, 1] = matrix[0, 5] + matrix[1, 5] + matrix[3, 4]
    stiffness_deviat[0, 2] = matrix[0, 4] + matrix[2, 4] + matrix[3, 5]
    stiffness_deviat[1, 2] = matrix[1, 3] + matrix[2, 3] + matrix[4, 5]
    stiffness_deviat = upper_tri_to_symmetric(stiffness_deviat)
    return stiffness_dilat, stiffness_deviat


@nb.njit(fastmath=True)
def mono_project(voigt_vector):
    """Project 21-component `voigt_vector` onto monoclinic symmetry subspace.

    Monoclinic symmetry is characterised by 13 independent elasticity components.

    See [Browaeys & Chevrot (2004)](https://doi.org/10.1111/j.1365-246X.2004.02415.x).

    """
    out = voigt_vector.copy()
    inds = (9, 10, 12, 13, 15, 16, 18, 19)
    for i in inds:
        out[i] = 0
    return out


@nb.njit(fastmath=True)
def ortho_project(voigt_vector):
    """Project 21-component `voigt_vector` onto orthorhombic symmetry subspace.

    Orthorhombic symmetry is characterised by 9 independent elasticity components.

    See [Browaeys & Chevrot (2004)](https://doi.org/10.1111/j.1365-246X.2004.02415.x).

    """
    out = voigt_vector.copy()
    out[9:] = 0
    return out


@nb.njit(fastmath=True)
def tetr_project(voigt_vector):
    """Project 21-component `voigt_vector` onto tetragonal symmetry subspace.

    Tetragonal symmetry is characterised by 6 independent elasticity components.

    See [Browaeys & Chevrot (2004)](https://doi.org/10.1111/j.1365-246X.2004.02415.x).

    """
    out = ortho_project(voigt_vector)
    for i, j in ((0, 1), (3, 4), (6, 7)):
        for k in range(2):
            out[i + k] = 0.5 * (voigt_vector[i] + voigt_vector[j])
    return out


@nb.njit(fastmath=True)
def hex_project(voigt_vector):
    """Project 21-component `voigt_vector` onto hexagonal symmetry subspace.

    Hexagonal symmetry (a.k.a. transverse isotropy) is characterised by 5 independent
    elasticity components.

    See [Browaeys & Chevrot (2004)](https://doi.org/10.1111/j.1365-246X.2004.02415.x).

    """
    x = voigt_vector
    out = np.zeros(21)
    out[0] = out[1] = 3 / 8 * (x[0] + x[1]) + x[5] / 4 / np.sqrt(2) + x[8] / 4
    out[2] = x[2]
    out[3] = out[4] = (x[3] + x[4]) / 2
    out[5] = (x[0] + x[1]) / 4 / np.sqrt(2) + 3 / 4 * x[5] - x[8] / 2 / np.sqrt(2)
    out[6] = out[7] = (x[6] + x[7]) / 2
    out[8] = (x[0] + x[1]) / 4 - x[5] / 2 / np.sqrt(2) + x[8] / 2
    return out


@nb.njit(fastmath=True)
def upper_tri_to_symmetric(arr):
    """Create symmetric array using upper triangle of input array.

    Examples:

    >>> import numpy as np
    >>> upper_tri_to_symmetric(np.array([
    ...         [ 1.,  2.,  3.,  4.],
    ...         [ 0.,  5.,  6.,  7.],
    ...         [ 0.,  0.,  8.,  9.],
    ...         [ 9.,  0.,  0., 10.]
    ... ]))
    array([[ 1.,  2.,  3.,  4.],
           [ 2.,  5.,  6.,  7.],
           [ 3.,  6.,  8.,  9.],
           [ 4.,  7.,  9., 10.]])

    """
    # <https://stackoverflow.com/questions/58718365/fast-way-to-convert-upper-triangular-matrix-into-symmetric-matrix>
    upper_tri = np.triu(arr)
    return np.where(upper_tri, upper_tri, upper_tri.transpose())


@nb.njit(fastmath=True)
def voigt_to_elastic_tensor(matrix):
    """Create 4-th order elastic tensor from an equivalent Voigt matrix.

    See also: `elastic_tensor_to_voigt`.

    """
    tensor = np.empty((3, 3, 3, 3))
    for p in range(3):
        for q in range(3):
            delta_pq = 1 if p == q else 0
            i = (p + 1) * delta_pq + (1 - delta_pq) * (7 - p - q) - 1
            for r in range(3):
                for s in range(3):
                    delta_rs = 1 if r == s else 0
                    j = (r + 1) * delta_rs + (1 - delta_rs) * (7 - r - s) - 1
                    tensor[p, q, r, s] = matrix[i, j]
    return tensor


@nb.njit(fastmath=True)
def elastic_tensor_to_voigt(tensor):
    """Create a 6x6 Voigt matrix from an equivalent 4-th order elastic tensor."""
    matrix = np.zeros((6, 6))
    matrix_indices = np.zeros((6, 6))
    for p in range(3):
        for q in range(3):
            delta_pq = 1 if p == q else 0
            i = (p + 1) * delta_pq + (1 - delta_pq) * (7 - p - q) - 1
            for r in range(3):
                for s in range(3):
                    delta_rs = 1 if r == s else 0
                    j = (r + 1) * delta_rs + (1 - delta_rs) * (7 - r - s) - 1
                    matrix[i, j] += tensor[p, q, r, s]
                    matrix_indices[i, j] += 1
    matrix /= matrix_indices
    return (matrix + matrix.transpose()) / 2


@nb.njit(fastmath=True)
def voigt_matrix_to_vector(matrix):
    """Create the 21-component Voigt vector equivalent to the 6x6 Voigt matrix."""
    vector = np.zeros(21)
    for i in range(3):
        vector[i] = matrix[i, i]
        vector[i + 3] = np.sqrt(2) * matrix[(i + 1) % 3, (i + 2) % 3]
        vector[i + 6] = 2 * matrix[i + 3, i + 3]
        vector[i + 9] = 2 * matrix[i, i + 3]
        vector[i + 12] = 2 * matrix[(i + 2) % 3, i + 3]
        vector[i + 15] = 2 * matrix[(i + 1) % 3, i + 3]
        vector[i + 18] = 2 * np.sqrt(2) * matrix[(i + 1) % 3 + 3, (i + 2) % 3 + 3]
    return vector


@nb.njit(fastmath=True)
def voigt_vector_to_matrix(vector):
    """Create the 6x6 matrix representation of the 21-component Voigt vector.

    See also: `voigt_matrix_to_vector`.

    """
    matrix = np.zeros((6, 6))
    for i in range(3):
        matrix[i, i] = vector[i]
        matrix[i + 3, i + 3] = 0.5 * vector[i + 6]

    matrix[1, 2] = 1 / np.sqrt(2) * vector[3]
    matrix[0, 2] = 1 / np.sqrt(2) * vector[4]
    matrix[0, 1] = 1 / np.sqrt(2) * vector[5]

    matrix[0, 3] = 0.5 * vector[9]
    matrix[1, 4] = 0.5 * vector[10]
    matrix[2, 5] = 0.5 * vector[11]
    matrix[2, 3] = 0.5 * vector[12]

    matrix[0, 4] = 0.5 * vector[13]
    matrix[1, 5] = 0.5 * vector[14]
    matrix[1, 3] = 0.5 * vector[15]

    matrix[2][4] = 0.5 * vector[16]
    matrix[0][5] = 0.5 * vector[17]
    matrix[4][5] = 0.5 * 1 / np.sqrt(2) * vector[18]
    matrix[3][5] = 0.5 * 1 / np.sqrt(2) * vector[19]
    matrix[3][4] = 0.5 * 1 / np.sqrt(2) * vector[20]
    return upper_tri_to_symmetric(matrix)


@nb.njit(fastmath=True)
def rotate(tensor, rotation):
    """Rotate 4-th order tensor using a 3x3 rotation matrix."""
    rotated_tensor = np.zeros((3, 3, 3, 3))
    for i in range(3):
        for j in range(3):
            for k in range(3):
                for L in range(3):
                    for a in range(3):
                        for b in range(3):
                            for c in range(3):
                                for d in range(3):
                                    rotated_tensor[i, j, k, L] += (
                                        rotation[i, a]
                                        * rotation[j, b]
                                        * rotation[k, c]
                                        * rotation[L, d]
                                        * tensor[a, b, c, d]
                                    )
    return rotated_tensor



"""> PyDRex: Mock objects for testing and reproducibility."""

from pydrex.core import MineralFabric

PARAMS_FRATERS2021 = {
    "olivine_fraction": 0.7,
    "enstatite_fraction": 0.3,
    "initial_olivine_fabric": MineralFabric.olivine_A,
    "stress_exponent": 1.5,
    "deformation_exponent": 3.5,
    "gbm_mobility": 125,
    "gbs_threshold": 0.3,
    "nucleation_efficiency": 5,
    "number_of_grains": 5000,
}
"""Values used for tests 1, 2 and 4 in <https://doi.org/10.1029/2021gc009846>."""

PARAMS_KAMINSKI2001_FIG5_SOLID = {
    "olivine_fraction": 1,
    "enstatite_fraction": 0,
    "initial_olivine_fabric": MineralFabric.olivine_A,
    "stress_exponent": 1.5,
    "deformation_exponent": 3.5,
    "gbm_mobility": 0,
    "gbs_threshold": 0,
    "nucleation_efficiency": 5,
    "number_of_grains": 3375,  # 15^3 as in the Fortran code.
}
"""Values used for the M*=0 test in <https://doi.org/10.1016/s0012-821x(01)00356-9>."""

PARAMS_KAMINSKI2001_FIG5_SHORTDASH = {
    "olivine_fraction": 1,
    "enstatite_fraction": 0,
    "initial_olivine_fabric": MineralFabric.olivine_A,
    "stress_exponent": 1.5,
    "deformation_exponent": 3.5,
    "gbm_mobility": 50,
    "gbs_threshold": 0,
    "nucleation_efficiency": 5,
    "number_of_grains": 3375,  # 15^3 as in the Fortran code.
}
"""Values used for the M*=50 test in <https://doi.org/10.1016/s0012-821x(01)00356-9>."""

PARAMS_KAMINSKI2001_FIG5_LONGDASH = {
    "olivine_fraction": 1,
    "enstatite_fraction": 0,
    "initial_olivine_fabric": MineralFabric.olivine_A,
    "stress_exponent": 1.5,
    "deformation_exponent": 3.5,
    "gbm_mobility": 200,
    "gbs_threshold": 0,
    "nucleation_efficiency": 5,
    "number_of_grains": 3375,  # 15^3 as in the Fortran code.
}
"""Values used for the M*=200 test in <https://doi.org/10.1016/s0012-821x(01)00356-9>."""

PARAMS_KAMINSKI2004_FIG4_TRIANGLES = {
    "olivine_fraction": 1,
    "enstatite_fraction": 0,
    "initial_olivine_fabric": MineralFabric.olivine_A,
    "stress_exponent": 1.5,
    "deformation_exponent": 3.5,
    "gbm_mobility": 125,
    "gbs_threshold": 0.4,
    "nucleation_efficiency": 5,
    "number_of_grains": 4394,  # Section 4.1, first paragraph.
}
"""Values used for the χ=0.4 test in <https://doi.org/10.1111/j.1365-246x.2004.02308.x>."""

PARAMS_KAMINSKI2004_FIG4_SQUARES = {
    "olivine_fraction": 1,
    "enstatite_fraction": 0,
    "initial_olivine_fabric": MineralFabric.olivine_A,
    "stress_exponent": 1.5,
    "deformation_exponent": 3.5,
    "gbm_mobility": 125,
    "gbs_threshold": 0.2,
    "nucleation_efficiency": 5,
    "number_of_grains": 4394,  # Section 4.1, first paragraph.
}
"""Values used for the χ=0.2 test in <https://doi.org/10.1111/j.1365-246x.2004.02308.x>."""

PARAMS_KAMINSKI2004_FIG4_CIRCLES = {
    "olivine_fraction": 1,
    "enstatite_fraction": 0,
    "initial_olivine_fabric": MineralFabric.olivine_A,
    "stress_exponent": 1.5,
    "deformation_exponent": 3.5,
    "gbm_mobility": 125,
    "gbs_threshold": 0,
    "nucleation_efficiency": 5,
    "number_of_grains": 4394,  # Section 4.1, first paragraph.
}
"""Values used for the χ=0 test in <https://doi.org/10.1111/j.1365-246x.2004.02308.x>."""

PARAMS_HEDJAZIAN2017 = {
    "olivine_fraction": 0.7,
    "enstatite_fraction": 0.3,
    "initial_olivine_fabric": MineralFabric.olivine_A,
    "stress_exponent": 1.5,
    "deformation_exponent": 3.5,
    "gbm_mobility": 10,
    "gbs_threshold": 0.2,
    "nucleation_efficiency": 5,
    "number_of_grains": 2197,  # 13^3 for both olivine and enstatite.
}
"""Values used for the MOR model in <https://doi.org/10.1016/j.epsl.2016.12.004>."""



"""> PyDRex: Core D-Rex functions and enums.

The function `derivatives` implements the core D-Rex solver, which computes the
crystallographic rotation rate and changes in fractional grain volumes.

**Acronyms:**
- CRSS = Critical Resolved Shear Stress,
    i.e. threshold stress required to initiate slip on a slip system,
    normalised to the stress required to initiate slip on the softest slip system

"""

from enum import IntEnum, unique

import numba as nb
import numpy as np

# NOTE: Do NOT import any pydrex submodules here to avoid cyclical imports.


_USE_ORIGINAL_DREX = False  # Switch to use D-Rex 2004 version without corrections.


PERMUTATION_SYMBOL = np.array(
    [
        [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, -1.0, 0.0]],
        [[0.0, 0.0, -1.0], [0.0, 0.0, 0.0], [1.0, 0.0, 0.0]],
        [[0.0, 1.0, 0.0], [-1.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
    ]
)


@unique
class MineralPhase(IntEnum):
    """Supported mineral phases."""

    olivine = 0
    enstatite = 1


@unique
class DeformationRegime(IntEnum):
    """Deformation mechanism regimes."""

    diffusion = 0
    dislocation = 1
    byerlee = 2
    max_viscosity = 3


@unique
class MineralFabric(IntEnum):
    """Supported mineral fabrics.

    The following fabric types are supported:
    - olivine A-E type fabrics according to e.g.
      [Karato et al. (2008)](https://doi.org/10.1146%2Fannurev.earth.36.031207.124120).
    - enstatite AB fabric, see
      [Bernard et al. (2021)](https://doi.org/10.1016/j.tecto.2021.228954).

    """

    olivine_A = 0
    olivine_B = 1
    olivine_C = 2
    olivine_D = 3
    olivine_E = 4
    enstatite_AB = 5


@nb.njit
def get_crss(phase, fabric):
    """Get Critical Resolved Shear Stress for the mineral `phase` and `fabric`.

    Returns an array of the normalised threshold stresses required to activate slip on
    each slip system. Olivine slip systems are ordered according to the convention used
    for `pydrex.minerals.OLIVINE_SLIP_SYSTEMS`.

    """
    if phase == MineralPhase.olivine:
        match fabric:
            case MineralFabric.olivine_A:
                return np.array([1, 2, 3, np.inf])
            case MineralFabric.olivine_B:
                return np.array([3, 2, 1, np.inf])
            case MineralFabric.olivine_C:
                return np.array([3, 2, np.inf, 1])
            case MineralFabric.olivine_D:
                return np.array([1, 1, 3, np.inf])
            case MineralFabric.olivine_E:
                return np.array([3, 1, 2, np.inf])
            case _:
                raise ValueError(f"unsupported olivine fabric: {fabric}")
    elif phase == MineralPhase.enstatite:
        if fabric == MineralFabric.enstatite_AB:
            return np.array([np.inf, np.inf, np.inf, 1])
        raise ValueError(f"unsupported enstatite fabric: {fabric}")
    raise ValueError(f"phase must be a valid `MineralPhase`, not {phase}")


# 12 args is a lot, but this way we can use numba
# (only primitives and numpy containers allowed).
@nb.njit(fastmath=True)
def derivatives(
    phase,
    fabric,
    n_grains,
    orientations,
    fractions,
    strain_rate,
    velocity_gradient,
    stress_exponent,
    deformation_exponent,
    nucleation_efficiency,
    gbm_mobility,
    volume_fraction,
):
    """Get derivatives of orientation and volume distribution.

    Args:
    - `phase` (`MineralPhase`) — ordinal number of the mineral phase
    - `fabric` (`MineralFabric`) — ordinal number of the fabric type
    - `n_grains` (int) — number of "grains" i.e. discrete volume segments
    - `orientations` (array) — `n_grains`x3x3 orientations (direction cosines)
    - `fractions` (array) — volume fractions of the grains relative to aggregate volume
    - `strain_rate` (array) — 3x3 dimensionless macroscopic strain-rate tensor
    - `velocity_gradient` (array) — 3x3 dimensionless macroscopic velocity gradient
    - `stress_exponent` (float) — `p` in `dislocation_density ∝ shear_stress^p`
    - `deformation_exponent` (float) — `n` in `shear_stress ∝ |deformation_rate|^(1/n)`
    - `nucleation_efficiency` (float) — parameter controlling grain nucleation
    - `gmb_mobility` (float) — grain boundary mobility parameter
    - `volume_fraction` (float) — volume fraction of the mineral phase relative to
                                  other phases (multiphase simulations)

    Returns a tuple with the rotation rates and grain volume fraction changes.

    """
    # Based on subroutine DERIV in original Fortran.
    strain_energies = np.empty(n_grains)
    orientations_diff = np.empty((n_grains, 3, 3))
    for grain_index in range(n_grains):
        orientation_change, strain_energy = _get_rotation_and_strain(
            phase,
            fabric,
            orientations[grain_index],
            strain_rate,
            velocity_gradient,
            stress_exponent,
            deformation_exponent,
            nucleation_efficiency,
        )
        orientations_diff[grain_index] = orientation_change
        strain_energies[grain_index] = strain_energy
    # Volume average mean strain energy.
    mean_energy = np.sum(fractions * strain_energies)
    # Strain energy residual.
    strain_residuals = mean_energy - strain_energies
    fractions_diff = volume_fraction * gbm_mobility * fractions * strain_residuals
    return orientations_diff, fractions_diff


@nb.njit(fastmath=True)
def _get_deformation_rate(phase, orientation, slip_rates):
    """Calculate deformation rate tensor for olivine or enstatite.

    Calculate the deformation rate with respect to the local coordinate frame,
    defined by the principal strain axes (finite strain ellipsoid).

    Args:
    - `phase` (`MineralPhase`) — ordinal number of the mineral phase
    - `orientation` (array) — 3x3 orientation matrix (direction cosines)
    - `slip_rates` (array) — slip rates relative to slip rate on softest slip system

    """
    # This is called the 'G'/'slip' tensor in the Fortran code, aka the Schmid tensor.
    deformation_rate = np.empty((3, 3))
    for i in range(3):
        for j in range(3):
            deformation_rate[i, j] = 2 * (
                slip_rates[0] * orientation[0, i] * orientation[1, j]
                + slip_rates[1] * orientation[0, i] * orientation[2, j]
                + slip_rates[2] * orientation[2, i] * orientation[1, j]
                + slip_rates[3] * orientation[2, i] * orientation[0, j]
            )
    return deformation_rate


@nb.njit(fastmath=True)
def _get_slip_rate_softest(deformation_rate, velocity_gradient):
    """Calculate dimensionless strain rate on the softest slip system.

    Args:
    - `deformation_rate` (array) — 3x3 dimensionless deformation rate matrix
    - `velocity_gradient` (array) — 3x3 dimensionless velocity gradient matrix

    """
    # See eq. 4 in Fraters 2021.
    enumerator = 0
    denominator = 0

    for j in range(3):
        # NOTE: Mistake in original DRex code (j + 2), see Fraters & Billen 2021 S1.
        if _USE_ORIGINAL_DREX:
            k = (j + 2) % 3
        else:
            k = (j + 1) % 3
        enumerator -= (velocity_gradient[j, k] - velocity_gradient[k, j]) * (
            deformation_rate[j, k] - deformation_rate[k, j]
        )
        # NOTE: Mistake in Kaminski 2001 eq. 7: kl+1 instead of kk+1
        # See Fraters & Billen 2021 supplementary informaton S1.
        denominator -= (deformation_rate[j, k] - deformation_rate[k, j]) ** 2

        for L in range(3):
            enumerator += 2 * deformation_rate[j, L] * velocity_gradient[j, L]
            denominator += 2 * deformation_rate[j, L] ** 2

    # Avoid zero divisions:
    # Tiny denominator means that relevant deformation_rate entries are zero.
    # No deformation rate means no slip rate.
    if -1e-15 < denominator < 1e-15:
        return 0.0
    return enumerator / denominator


@nb.njit(fastmath=True)
def _get_slip_rates_olivine(invariants, slip_indices, crss, deformation_exponent):
    """Calculate relative slip rates of the active slip systems for olivine.

    Args:
    - `invariants` (array) — strain rate invariants for the four slip systems
    - `slip_indices` (array) — indices that sort the CRSS by increasing slip-rate
                               activity
    - `crss` (array) — reference resolved shear stresses (CRSS), see `pydrex.fabric`
    - `deformation_exponent` (float) — `n` in `shear_stress ∝ |deformation_rate|^(1/n)`

    """
    i_inac, i_min, i_int, i_max = slip_indices
    # Ratio of slip rates on each slip system to slip rate on softest slip system.
    # Softest slip system has max. slip rate (aka activity).
    # See eq. 5, Kaminski 2001.
    prefactor = crss[i_max] / invariants[i_max]
    ratio_min = prefactor * invariants[i_min] / crss[i_min]
    ratio_int = prefactor * invariants[i_int] / crss[i_int]
    slip_rates = np.empty(4)
    slip_rates[i_inac] = 0  # Hardest system is completely inactive in olivine.
    slip_rates[i_min] = ratio_min * np.abs(ratio_min) ** (deformation_exponent - 1)
    slip_rates[i_int] = ratio_int * np.abs(ratio_int) ** (deformation_exponent - 1)
    slip_rates[i_max] = 1
    return slip_rates


@nb.njit(fastmath=True)
def _get_slip_invariants(strain_rate, orientation):
    r"""Calculate strain rate invariants for minerals with four slip systems.

    Calculates $I = ∑_{ij} l_{i} n_{j} \dot{ε}_{ij}$ for each slip sytem of:
    - (010)[100]
    - (001)[100]
    - (010)[001]
    - (100)[001]
    Only the last return value is relevant for enstatite.
    These are not configurable for now.

    Args:
    - `strain_rate` (array) — 3x3 dimensionless strain rate matrix
    - `orientation` (array) — 3x3 orientation matrix (direction cosines)

    """
    invariants = np.zeros(4)
    for i in range(3):
        for j in range(3):
            # (010)[100]
            invariants[0] += strain_rate[i, j] * orientation[0, i] * orientation[1, j]
            # (001)[100]
            invariants[1] += strain_rate[i, j] * orientation[0, i] * orientation[2, j]
            # (010)[001]
            invariants[2] += strain_rate[i, j] * orientation[2, i] * orientation[1, j]
            # (100)[001]
            invariants[3] += strain_rate[i, j] * orientation[2, i] * orientation[0, j]
    return invariants


@nb.njit(fastmath=True)
def _get_orientation_change(
    orientation, velocity_gradient, deformation_rate, slip_rate_softest
):
    """Calculate the rotation rate for a grain undergoing dislocation creep.

    Args:
    - `orientation` (array) — 3x3 orientation matrix (direction cosines)
    - `velocity_gradient` (array) — 3x3 dimensionless velocity gradient matrix
    - `deformation_rate` (float) — 3x3 dimensionless deformation rate matrix
    - `slip_rate_softest` (float) — slip rate on the softest (most active) slip system

    """
    orientation_change = np.zeros((3, 3))
    # Spin vector for the grain, see eq. 3 in Fraters 2021.
    # Includes rigid body rotation as well as the plastic deformation contribution.
    # This means that even with no active slip systems, the rotation will be nonzero
    # if there is a rotational (i.e. antisymmetric) component of the velocity_gradient.
    spin_vector = np.empty(3)
    for j in range(3):
        r = (j + 1) % 3
        s = (j + 2) % 3
        spin_vector[j] = (
            (velocity_gradient[s, r] - velocity_gradient[r, s])
            - (deformation_rate[s, r] - deformation_rate[r, s]) * slip_rate_softest
        ) / 2

    # Calculate rotation rate, see eq. 9 Kaminski & Ribe (2001).
    # Equivalent to:
    #   spin_matrix = np.einsum("ikj,k->ij", PERMUTATION_SYMBOL, spin_vector)
    #   orientation_change = spin_matrix.transpose() @ orientation
    for p in range(3):
        for q in range(3):
            for r in range(3):
                for s in range(3):
                    orientation_change[p, q] += (
                        PERMUTATION_SYMBOL[q, r, s] * orientation[p, s] * spin_vector[r]
                    )

    return orientation_change


@nb.njit(fastmath=True)
def _get_strain_energy(
    crss,
    slip_rates,
    slip_indices,
    slip_rate_softest,
    stress_exponent,
    deformation_exponent,
    nucleation_efficiency,
):
    """Calculate strain energy due to dislocations for an olivine grain.

    Args:
    - `crss` (array) — reference resolved shear stresses (CRSS), see `pydrex.fabric`
    - `slip_rates` (array) — slip rates relative to slip rate on softest slip system
    - `slip_indices` (array) — indices that sort the CRSS by increasing slip-rate
                               activity
    - `slip_rate_softest` (float) — slip rate on the softest (most active) slip system
    - `stress_exponent` (float) — `p` in `dislocation_density ∝ shear_stress^p`
    - `deformation_exponent` (float) — `n` in `shear_stress ∝ |deformation_rate|^(1/n)`
    - `nucleation_efficiency` (float) — parameter controlling grain nucleation

    Note that "new" grains are assumed to rotate with their parent.

    """
    strain_energy = 0.0
    # Dimensionless dislocation density for each slip system.
    # See eq. 16 Fraters 2021.
    # NOTE: Mistake in eq. 11, Kaminski 2004: spurious division by strain rate scale.
    # NOTE: Here we call 'p' the 'stress_exponent' and 'n' the 'deformation_exponent',
    # but in the original they use the variable 'stress_exponent' for 'n' (3.5).
    for i in range(3):
        dislocation_density = (1 / crss[i]) ** (
            deformation_exponent - stress_exponent
        ) * np.abs(slip_rates[i] * slip_rate_softest) ** (
            stress_exponent / deformation_exponent
        )
        # Dimensionless strain energy for this grain, see eq. 14, Fraters 2021.
        strain_energy += dislocation_density * np.exp(
            -nucleation_efficiency * dislocation_density**2
        )
    return strain_energy


@nb.njit(fastmath=True)
def _get_rotation_and_strain(
    phase,
    fabric,
    orientation,
    strain_rate,
    velocity_gradient,
    stress_exponent,
    deformation_exponent,
    nucleation_efficiency,
):
    """Get the crystal axes rotation rate and strain energy of individual grain.

    Args:
    - `phase` (`MineralPhase`) — ordinal number of the mineral phase
    - `fabric` (`MineralFabric`) — ordinal number of the fabric type
    - `orientation` (array) — 3x3 orientation matrix (direction cosines)
    - `strain_rate` (array) — 3x3 dimensionless strain rate matrix
    - `velocity_gradient` (array) — 3x3 dimensionless velocity gradient matrix
    - `stress_exponent` (float) — `p` in `dislocation_density ∝ shear_stress^p`
    - `deformation_exponent` (float) — `n` in `shear_stress ∝ |deformation_rate|^(1/n)`
    - `nucleation_efficiency (float) — parameter controlling grain nucleation

    Note that "new" grains are assumed to rotate with their parent.

    Returns a tuple with the rotation rate of the crystalline axes
    with respect to the principal strain axes and strain energy of the grain.

    """
    crss = get_crss(phase, fabric)
    slip_invariants = _get_slip_invariants(strain_rate, orientation)
    if phase == MineralPhase.olivine:
        slip_indices = np.argsort(np.abs(slip_invariants / crss))
        slip_rates = _get_slip_rates_olivine(
            slip_invariants,
            slip_indices,
            crss,
            deformation_exponent,
        )
    elif phase == MineralPhase.enstatite:
        slip_indices = np.argsort(1 / crss)
        slip_rates = np.zeros(4)
        if _USE_ORIGINAL_DREX:
            slip_rates[-1] = 1  # Original had an arbitrary slip always active (L1410).
        else:
            # Assumes exclusively (100)[001] slip for enstatite.
            if np.abs(slip_invariants[-1]) > 1e-15:
                slip_rates[-1] = 1
    else:
        assert False  # Should never happen.

    deformation_rate = _get_deformation_rate(phase, orientation, slip_rates)
    slip_rate_softest = _get_slip_rate_softest(deformation_rate, velocity_gradient)
    orientation_change = _get_orientation_change(
        orientation,
        velocity_gradient,
        deformation_rate,
        slip_rate_softest,
    )
    strain_energy = _get_strain_energy(
        crss,
        slip_rates,
        slip_indices,
        slip_rate_softest,
        stress_exponent,
        deformation_exponent,
        nucleation_efficiency,
    )
    return orientation_change, strain_energy



"""> PyDRex: Visualisation functions for test outputs and examples."""

import numpy as np
from cmcrameri import cm as cmc
from matplotlib import projections as mproj
from matplotlib import pyplot as plt

from pydrex import axes as _axes
from pydrex import core as _core
from pydrex import io as _io
from pydrex import logger as _log
from pydrex import utils as _utils

# Get default figure size for easy referencing and scaling.
DEFAULT_FIG_WIDTH, DEFAULT_FIG_HEIGHT = plt.rcParams["figure.figsize"]
plt.rcParams["axes.grid"] = True
# Always draw grid behind everything else.
plt.rcParams["axes.axisbelow"] = True
# Always use constrained layout by default (modern version of tight layout).
plt.rcParams["figure.constrained_layout.use"] = True
# Use 300 DPI by default, NASA can keep their blurry images.
plt.rcParams["figure.dpi"] = 300
# Make sure we have the required matplotlib "projections" (really just Axes subclasses).
if "pydrex.polefigure" not in mproj.get_projection_names():
    _log.warning(
        "failed to find pydrex.polefigure projection; it should be registered in %s",
        _axes,
    )


def polefigures(
    orientations,
    ref_axes,
    i_range,
    density=False,
    savefile="polefigures.png",
    strains=None,
    **kwargs,
):
    """Plot pole figures of a series of (Nx3x3) orientation matrix stacks.

    Produces [100], [010] and [001] pole figures for (resampled) orientations.
    For the argument specification, check the output of `pydrex-polefigures --help`
    on the command line.

    """
    if len(orientations) != len(i_range):
        raise ValueError("mismatched length of 'orientations' and 'i_range'")
    if strains is not None and len(strains) != len(i_range):
        raise ValueError("mismatched length of 'strains'")
    n_orientations = len(orientations)
    fig = plt.figure(figsize=(n_orientations, 4), dpi=600)

    if len(i_range) == 1:
        grid = fig.add_gridspec(3, n_orientations, hspace=0, wspace=0.2)
        first_row = 0
    else:
        grid = fig.add_gridspec(
            4, n_orientations, height_ratios=((1, 3, 3, 3)), hspace=0, wspace=0.2
        )
        fig_strain = fig.add_subfigure(grid[0, :])
        first_row = 1
        ax_strain = fig_strain.add_subplot(111)

        if strains is None:
            fig_strain.suptitle(
                f"N ⋅ (max strain) / {i_range.stop}", x=0.5, y=0.85, fontsize="small"
            )
            ax_strain.set_xlim(
                (i_range.start - i_range.step / 2, i_range.stop - i_range.step / 2)
            )
            ax_strain.set_xticks(list(i_range))
        else:
            fig_strain.suptitle("strain (%)", x=0.5, y=0.85, fontsize="small")
            ax_strain.set_xticks(strains)
            ax_strain.set_xlim(
                (
                    strains[0] - strains[1] / 2,
                    strains[-1] + strains[1] / 2,
                )
            )

        ax_strain.set_frame_on(False)
        ax_strain.grid(False)
        ax_strain.yaxis.set_visible(False)
        ax_strain.xaxis.set_tick_params(labelsize="x-small", length=0)

    fig100 = fig.add_subfigure(
        grid[first_row, :], edgecolor=plt.rcParams["grid.color"], linewidth=1
    )
    fig100.suptitle("[100]", fontsize="small")
    fig010 = fig.add_subfigure(
        grid[first_row + 1, :], edgecolor=plt.rcParams["grid.color"], linewidth=1
    )
    fig010.suptitle("[010]", fontsize="small")
    fig001 = fig.add_subfigure(
        grid[first_row + 2, :], edgecolor=plt.rcParams["grid.color"], linewidth=1
    )
    fig001.suptitle("[001]", fontsize="small")
    for n, orientations in enumerate(orientations):
        ax100 = fig100.add_subplot(
            1, n_orientations, n + 1, projection="pydrex.polefigure"
        )
        pf100 = ax100.polefigure(
            orientations,
            hkl=[1, 0, 0],
            ref_axes=ref_axes,
            density=density,
            density_kwargs=kwargs,
        )
        ax010 = fig010.add_subplot(
            1, n_orientations, n + 1, projection="pydrex.polefigure"
        )
        pf010 = ax010.polefigure(
            orientations,
            hkl=[0, 1, 0],
            ref_axes=ref_axes,
            density=density,
            density_kwargs=kwargs,
        )
        ax001 = fig001.add_subplot(
            1, n_orientations, n + 1, projection="pydrex.polefigure"
        )
        pf001 = ax001.polefigure(
            orientations,
            hkl=[0, 0, 1],
            ref_axes=ref_axes,
            density=density,
            density_kwargs=kwargs,
        )
        if density:
            for ax, pf in zip(
                (ax100, ax010, ax001), (pf100, pf010, pf001), strict=True
            ):
                cbar = fig.colorbar(
                    pf,
                    ax=ax,
                    fraction=0.05,
                    location="bottom",
                    orientation="horizontal",
                )
                cbar.ax.xaxis.set_tick_params(labelsize="xx-small")

    fig.savefig(_io.resolve_path(savefile))


def pathline_box2d(
    ax,
    get_velocity,
    ref_axes,
    colors,
    positions,
    marker,
    min_coords,
    max_coords,
    resolution,
    aspect="equal",
    cmap=cmc.batlow,
    cpo_vectors=None,
    cpo_strengths=None,
    tick_formatter=lambda x, pos: f"{x/1e3:.1f}",
    **kwargs,
):
    """Plot pathlines and velocity arrows for a 2D box domain.

    If `ax` is None, a new figure and axes are created with `figure_unless`.

    Args:
    - `get_velocity` (callable) — callable with signature f(t, x) that returns the 3D
      velocity vector at a given time (not used) and 3D position vector
    - `ref_axes` (two letters from {"x", "y", "z"}) — labels for the horizontal and
      vertical axes (these also define the projection for the 3D velocity/position)
    - `colors` (array) — monotonic values along a representative pathline in the flow
    - `positions` (Nx3 array) — 3D position vectors along the same pathline
    - `min_coords` (array) — 2D coordinates of the lower left corner of the domain
    - `max_coords` (array) — 2D coordinates of the upper right corner of the domain
    - `resolution` (array) — 2D resolution of the velocity arrow grid (i.e. number of
      grid points in the horizontal and vertical directions) which can be set to None to
      prevent drawing velocity vectors
    - `aspect` (str|float, optional) — see `matplotlib.axes.Axes.set_aspect`
    - `cmap` (Matplotlib color map, optional) — color map for `colors`
    - `cpo_vectors` (array, optional) — vectors to plot as bars at pathline locations
    - `cpo_strengths` (array, optional) — strengths used to scale the cpo bars
    - `tick_formatter` (callable, optional) — function used to format tick labels

    Additional keyword arguments are passed to the `matplotlib.axes.Axes.quiver` call
    used to plot the velocity vectors.

    Returns the figure handle, the axes handle, the quiver collection (velocities) and
    the scatter collection (pathline).

    """
    fig, ax = figure_unless(ax)
    ax.set_xlabel(ref_axes[0])
    ax.set_ylabel(ref_axes[1])

    x_min, y_min = min_coords
    x_max, y_max = max_coords
    ax.set_xlim((x_min, x_max))
    ax.set_ylim((y_min, y_max))
    ax.set_aspect(aspect)
    ax.xaxis.set_major_formatter(tick_formatter)
    ax.yaxis.set_major_formatter(tick_formatter)

    _ref_axes = ref_axes.lower()
    axes_map = {"x": 0, "y": 1, "z": 2}
    horizontal = axes_map[_ref_axes[0]]
    vertical = axes_map[_ref_axes[1]]

    velocities = None
    if resolution is not None:
        x_res, y_res = resolution
        X = np.linspace(x_min, x_max, x_res)
        Y = np.linspace(y_min, y_max, y_res)
        X_grid, Y_grid = np.meshgrid(X, Y)

        U = np.zeros_like(X_grid.ravel())
        V = np.zeros_like(Y_grid.ravel())
        for i, (x, y) in enumerate(zip(X_grid.ravel(), Y_grid.ravel(), strict=True)):
            p = np.zeros(3)
            p[horizontal] = x
            p[vertical] = y
            v3d = get_velocity(np.nan, p)
            U[i] = v3d[horizontal]
            V[i] = v3d[vertical]

        velocities = ax.quiver(
            X_grid,
            Y_grid,
            U.reshape(X_grid.shape),
            V.reshape(Y_grid.shape),
            pivot=kwargs.pop("pivot", "mid"),
            alpha=kwargs.pop("alpha", 0.25),
            **kwargs,
        )

    P = np.asarray([[p[horizontal], p[vertical]] for p in positions])
    if cpo_vectors is not None:
        if cpo_strengths is None:
            cpo_strengths = np.full(len(cpo_vectors), 1.0)
        C = np.asarray(
            [
                f * np.asarray([c[horizontal], c[vertical]])
                for f, c in zip(cpo_strengths, cpo_vectors, strict=True)
            ]
        )
        cpo = ax.quiver(
            P[:, 0],
            P[:, 1],
            C[:, 0],
            C[:, 1],
            colors,
            cmap=cmap,
            pivot="mid",
            width=kwargs.pop("width", 3e-3),
            headaxislength=0,
            headlength=0,
            zorder=kwargs.pop("zorder", 10) + 1,  # Always above velocity vectors.
        )
    else:
        cpo = ax.scatter(P[:, 0], P[:, 1], marker=marker, c=colors, cmap=cmap)
    return fig, ax, velocities, cpo


def alignment(
    ax,
    strains,
    angles,
    markers,
    labels,
    err=None,
    θ_max=90,
    θ_fse=None,
    colors=None,
    cmaps=None,
    **kwargs,
):
    """Plot `angles` (in degrees) versus `strains` on the given axis.

    Alignment angles could be either bingham averages or the a-axis in the hexagonal
    symmetry projection, measured from e.g. the shear direction. In the first case,
    they should be calculated from resampled grain orientations. Expects as many
    `markers` and `labels` as there are data series in `angles`.

    If `ax` is None, a new figure and axes are created with `figure_unless`.

    Args:
    - `strains` (array) — X-values, accumulated strain (tensorial) during CPO evolution,
      may be a 2D array of multiple strain series
    - `angles` (array) — Y-values, may be a 2D array of multiple angle series
    - `markers` (sequence) — MatPlotLib markers to use for the data series
    - `labels` (sequence) — labels to use for the data series
    - `err` (array, optional) — standard errors for the `angles`, shapes must match
    - `θ_max` (int) — maximum angle (°) to show on the plot, should be less than 90
    - `θ_fse` (array, optional) — an array of angles from the long axis of the finite
      strain ellipsoid to the reference direction (e.g. shear direction)
    - `colors` (array, optional) — color coordinates for series of angles
    - `cmaps` (Matplotlib color maps, optional) — color maps for `colors`

    If `colors` and `cmaps` are used, then angle values are colored individually within
    each angle series.

    Additional keyword arguments are passed to `matplotlib.axes.Axes.scatter` if
    `colors` is not `None`, or to `matplotlib.axes.Axes.plot` otherwise.

    Returns a tuple of the figure handle, the axes handle and the set of colors used for
    the data series plots.

    """
    _strains = np.atleast_2d(strains)
    _angles = np.atleast_2d(angles)
    if err is not None:
        _angles_err = np.atleast_2d(err)
    if not np.all(_strains.shape == _angles.shape):
        # Assume strains are all the same for each series in `angles`, try np.tile().
        _strains = np.tile(_strains, (len(_angles), 1))

    fig, ax = figure_unless(ax)
    ax.set_ylabel(r"$\overline{θ}$ ∈ [0, 90]°")
    ax.set_ylim((0, θ_max))
    ax.set_xlabel("Strain (ε)")
    ax.set_xlim((np.min(strains), np.max(strains)))
    _colors = []
    for i, (strains, θ_cpo, marker, label) in enumerate(
        zip(_strains, _angles, markers, labels, strict=True)
    ):
        if colors is not None:
            ax.scatter(
                strains,
                θ_cpo,
                marker=marker,
                label=label,
                c=colors[i],
                cmap=cmaps[i],
                alpha=kwargs.pop("alpha", 0.6),
                edgecolor=kwargs.pop("edgecolor", plt.rcParams["axes.edgecolor"]),
                **kwargs,
            )
            _colors.append(colors[i])
        else:
            lines = ax.plot(strains, θ_cpo, marker, alpha=0.6, label=label, **kwargs)
            _colors.append(lines[0].get_color())
        if err is not None:
            ax.fill_between(
                strains,
                θ_cpo - _angles_err[i],
                θ_cpo + _angles_err[i],
                alpha=0.22,
                color=_colors[i],
            )

    if θ_fse is not None:
        ax.plot(strains, θ_fse, linestyle=(0, (5, 5)), alpha=0.6, label="FSE")
    if not all(b is None for b in labels):
        _utils.redraw_legend(ax)
    return fig, ax, _colors


def strengths(
    ax,
    strains,
    strengths,
    ylabel,
    markers,
    labels,
    err=None,
    cpo_threshold=None,
    colors=None,
    cmaps=None,
    **kwargs,
):
    """Plot CPO `strengths` (e.g. M-indices) versus `strains` on the given axis.

    If `ax` is None, a new figure and axes are created with `figure_unless`.

    Args:
    - `strains` (array) — X-values, accumulated strain (tensorial) during CPO evolution,
      may be a 2D array of multiple strain series
    - `strengths` (array) — Y-values, may be a 2D array of multiple strength series
    - `markers` (sequence) — MatPlotLib markers to use for the data series
    - `labels` (sequence) — labels to use for the data series
    - `err` (array, optional) — standard errors for the `strengths`, shapes must match
    - `colors` (array, optional) — color coordinates for series of strengths
    - `cpo_threshold` (float, optional) — plot a dashed line at this threshold
    - `cmaps` (Matplotlib color maps, optional) — color maps for `colors`

    If `colors` and `cmaps` are used, then strength values are colored individually
    within each strength series.

    Additional keyword arguments are passed to `matplotlib.axes.Axes.scatter` if
    `colors` is not `None`, or to `matplotlib.axes.Axes.plot` otherwise.

    Returns a tuple of the figure handle, the axes handle and the set of colors used for
    the data series plots.

    """
    _strains = np.atleast_2d(strains)
    _strengths = np.atleast_2d(strengths)
    if err is not None:
        _strengths_err = np.atleast_2d(err)
    if not np.all(_strains.shape == _strengths.shape):
        # Assume strains are all the same for each series in `strengths`, try np.tile().
        _strains = np.tile(_strains, (len(_strengths), 1))

    fig, ax = figure_unless(ax)
    ax.set_ylabel(ylabel)
    ax.set_xlabel("Strain (ε)")
    ax.set_xlim((np.min(strains), np.max(strains)))

    if cpo_threshold is not None:
        ax.axhline(cpo_threshold, color=plt.rcParams["axes.edgecolor"], linestyle="--")

    _colors = []
    for i, (strains, strength, marker, label) in enumerate(
        zip(_strains, _strengths, markers, labels, strict=True)
    ):
        if colors is not None:
            ax.scatter(
                strains,
                strength,
                marker=marker,
                label=label,
                c=colors[i],
                cmap=cmaps[i],
                alpha=kwargs.pop("alpha", 0.6),
                edgecolor=kwargs.pop("edgecolor", plt.rcParams["axes.edgecolor"]),
                **kwargs,
            )
            _colors.append(colors[i])
        else:
            lines = ax.plot(
                strains, strength, marker, alpha=0.33, label=label, **kwargs
            )
            _colors.append(lines[0].get_color())
        if err is not None:
            ax.fill_between(
                strains,
                strength - _strengths_err[i],
                strength + _strengths_err[i],
                alpha=0.22,
                color=_colors[i],
            )

    if not all(b is None for b in labels):
        _utils.redraw_legend(ax)
    return fig, ax, _colors


def grainsizes(ax, strains, fractions):
    """Plot grain volume `fractions` versus `strains` on the given axis.

    If `ax` is None, a new figure and axes are created with `figure_unless`.

    """
    n_grains = len(fractions[0])
    fig, ax = figure_unless(ax)
    ax.set_ylabel(r"$\log_{10}(f × N)$")
    ax.set_xlabel("Strain (ε)")
    parts = ax.violinplot(
        [np.log10(f * n_grains) for f in fractions], positions=strains, widths=0.8
    )
    for part in parts["bodies"]:
        part.set_color("black")
        part.set_alpha(1)
    parts["cbars"].set_alpha(0)
    parts["cmins"].set_visible(False)
    parts["cmaxes"].set_visible(False)
    # parts["cmaxes"].set_color("red")
    # parts["cmaxes"].set_alpha(0.5)
    return fig, ax, parts


def show_Skemer2016_ShearStrainAngles(
    ax, studies, markers, colors, fillstyles, labels, fabric
):
    """Show data from `src/pydrex/data/thirdparty/Skemer2016_ShearStrainAngles.scsv`.

    Plot data from the Skemer 2016 datafile on the axis given by `ax`. Select the
    studies from which to plot the data, which must be a list of strings with exact
    matches in the `study` column in the datafile.
    Also filter the data to select only the given `fabric`
    (see `pydrex.core.MineralFabric`).

    If `ax` is None, a new figure and axes are created with `figure_unless`.

    Returns a tuple containing:
    - the figure handle
    - the axes handle
    - the set of colors used for the data series plots
    - the Skemer 2016 dataset
    - the indices used to select data according to the "studies" and "fabric" filters

    """
    fabric_map = {
        _core.MineralFabric.olivine_A: "A",
        _core.MineralFabric.olivine_B: "B",
        _core.MineralFabric.olivine_C: "C",
        _core.MineralFabric.olivine_D: "D",
        _core.MineralFabric.olivine_E: "E",
    }
    fig, ax = figure_unless(ax)
    data_Skemer2016 = _io.read_scsv(
        _io.data("thirdparty") / "Skemer2016_ShearStrainAngles.scsv"
    )
    for study, marker, color, fillstyle, label in zip(
        studies, markers, colors, fillstyles, labels, strict=True
    ):
        # Note: np.nonzero returns a tuple.
        indices = np.nonzero(
            np.logical_and(
                np.asarray(data_Skemer2016.study) == study,
                np.asarray(data_Skemer2016.fabric) == fabric_map[fabric],
            )
        )[0]
        ax.plot(
            np.take(data_Skemer2016.shear_strain, indices) / 200,
            np.take(data_Skemer2016.angle, indices),
            marker=marker,
            fillstyle=fillstyle,
            linestyle="none",
            color=color,
            label=label,
        )
    if not all(b is None for b in labels):
        _utils.redraw_legend(ax)
    return fig, ax, colors, data_Skemer2016, indices


def spin(
    ax,
    initial_angles,
    rotation_rates,
    target_initial_angles=None,
    target_rotation_rates=None,
    labels=("target", "computed"),
    shear_axis=None,
):
    """Plot rotation rates of grains with known, unique initial [100] angles from X.

    If `ax` is None, a new figure and axes are created with `figure_unless`.
    The default labels ("target", "computed") can also be overriden.
    If `shear_axis` is not None, a dashed line will be drawn at the given x-value
    (and its reflection around 180°).

    Returns a tuple of the figure handle, the axes handle and the set of colors used for
    the data series plots.

    """
    fig, ax = figure_unless(ax)
    ax.set_ylabel("Rotation rate (°/s)")
    ax.set_xlabel("Initial [100] angle (°)")
    ax.set_xlim((0, 360))
    ax.set_xticks(np.linspace(0, 360, 9))
    if shear_axis is not None:
        ax.axvline(shear_axis, color="k", linestyle="--", alpha=0.5)
        ax.axvline(
            (shear_axis + 180) % 360,
            color="k",
            linestyle="--",
            alpha=0.5,
            label="shear axis",
        )
    colors = []
    if target_rotation_rates is not None:
        lines = ax.plot(
            target_initial_angles,
            target_rotation_rates,
            c="tab:orange",
            label=labels[0],
        )
        colors.append(lines[0].get_color())
    series = ax.scatter(
        initial_angles,
        rotation_rates,
        facecolors="none",
        edgecolors=plt.rcParams["axes.edgecolor"],
        label=labels[1],
    )
    colors.append(series.get_edgecolors()[0])
    _utils.redraw_legend(ax)
    return fig, ax, colors


def growth(
    ax,
    initial_angles,
    fractions_diff,
    target_initial_angles=None,
    target_fractions_diff=None,
    labels=("target", "computed"),
    shear_axis=None,
):
    """Plot grain growth of grains with known, unique initial [100] angles from X.

    If `ax` is None, a new figure and axes are created with `figure_unless`.
    The default labels ("target", "computed") can also be overriden.
    If `shear_axis` is not None, a dashed line will be drawn at the given x-value
    (and its reflection around 180°).

    Returns a tuple of the figure handle, the axes handle and the set of colors used for
    the data series plots.

    """
    fig, ax = figure_unless(ax)
    ax.set_ylabel("Grain growth rate (s⁻¹)")
    ax.set_xlabel("Initial [100] angle (°)")
    ax.set_xlim((0, 360))
    ax.set_xticks(np.linspace(0, 360, 9))
    if shear_axis is not None:
        ax.axvline(shear_axis, color="k", linestyle="--", alpha=0.5)
        ax.axvline(
            (shear_axis + 180) % 360,
            color="k",
            linestyle="--",
            alpha=0.5,
            label="shear axis",
        )
    colors = []
    if target_fractions_diff is not None:
        lines = ax.plot(
            target_initial_angles,
            target_fractions_diff,
            c="tab:orange",
            label=labels[0],
        )
        colors.append(lines[0].get_color())
    series = ax.scatter(
        initial_angles,
        fractions_diff,
        facecolors="none",
        edgecolors=plt.rcParams["axes.edgecolor"],
        label=labels[1],
    )
    colors.append(series.get_edgecolors()[0])
    _utils.redraw_legend(ax)
    return fig, ax, colors


def figure_unless(ax):
    """Create figure and axes if `ax` is None, or return existing figure for `ax`.

    If `ax` is None, a new figure is created for the axes with a few opinionated default
    settings (grid, constrained layout, high DPI).

    Returns a tuple containing the figure handle and the axes object.

    """
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot()
    else:
        fig = ax.get_figure()
    return fig, ax


def figure(figscale=None, **kwargs):
    """Create new figure with a few opinionated default settings.

    (e.g. grid, constrained layout, high DPI).

    The keyword argument `figscale` can be used to scale the figure width and height
    relative to the default values by passing a tuple. Any additional keyword arguments
    are passed to `matplotlib.pyplot.figure()`.

    """
    # NOTE: Opinionated defaults are set using rcParams at the top of this file.
    _figsize = kwargs.pop("figsize", (DEFAULT_FIG_WIDTH, DEFAULT_FIG_HEIGHT))
    if figscale is not None:
        _figsize = (DEFAULT_FIG_WIDTH * figscale[0], DEFAULT_FIG_HEIGHT * figscale[1])
    return plt.figure(figsize=_figsize, **kwargs)
